/*
								+---------------------------------+
								|                                 |
								|    ***   Shader system   ***    |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|          ShaderSys.cpp          |
								|                                 |
								+---------------------------------+
*/

#include <vector>
#include "../berLame_src/CallStack.h"
#include <windows.h>
#include <gl/gl.h>
#include <gl/glext.h>
#include <gl/wglext.h>
#include <algorithm>
#include <functional>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLState.h"
#include "Texture.h"
#include "RenderBuffer2.h"
#include "Shader2.h"
#include "ShaderSys.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

/*
 *								=== TShaderInfo ===
 */

/*
 *	bool TShaderInfo::Dynamize()
 *		- presumably, the structure was filled by copying from another one or by const data
 *		- this function allocates new buffers for shader info and, copies current values
 *		  inside and then assigns those new buffers to be used by this shader info
 *		- returns true on success, false on failure (presumably not enough memory)
 */
bool TShaderInfo::Dynamize()
{
	const TShader *p_high_src = p_high;
	const TProgram *p_low_src = p_low;

	p_high = 0;
	p_low = 0;
	// it doesn't belong to us, in case this function fails, we don't want
	// to have memory that doesn't belong to us

	if(!(p_s_name = p_s_Alloc(p_s_name)))
		return false;
	// copy name

	if(p_high_src) {
		if(!(p_high = new TShader))
			return false;

		p_high->n_sampler_num = p_high_src->n_sampler_num;
		if(!(p_high->p_sampler = new TShader::TSamplerInfo[p_high->n_sampler_num]) ||
		   !(p_high->p_s_source_code = p_s_Alloc(p_high_src->p_s_source_code)))
			return false;
		// copy source code

		const TShader::TSamplerInfo *p_src_sampler = p_high_src->p_sampler;
		for(TShader::TSamplerInfo *p_sampler = p_high->p_sampler,
		   *p_end = p_high->p_sampler + p_high->n_sampler_num; p_sampler < p_end;
		   ++ p_sampler, ++ p_src_sampler) {
			if(!(p_sampler->p_s_name = p_s_Alloc(p_src_sampler->p_s_name)))
				return false;
			p_sampler->n_resource_index = p_src_sampler->n_resource_index;
		}
		// copy samplers
	}
	// copy "high" section

	if(p_low_src) {
		if(!(p_low = new TProgram))
			return false;

		p_low->n_texture_unit_num = p_low_src->n_texture_unit_num;
		p_low->n_parameter_num = p_low_src->n_parameter_num;
		if(!(p_low->p_texture_unit = new TProgram::TTexUnitInfo[p_low->n_texture_unit_num]) ||
		   !(p_low->p_parameter = new TProgram::TParamInfo[p_low->n_parameter_num]) ||
		   !(p_low->p_s_source_code = p_s_Alloc(p_low_src->p_s_source_code)))
			return false;
		// copy source code

		const TProgram::TParamInfo *p_src_parameter = p_low_src->p_parameter;
		for(TProgram::TParamInfo *p_parameter = p_low->p_parameter,
		   *p_end = p_low->p_parameter + p_low->n_parameter_num; p_parameter < p_end;
		   ++ p_parameter, ++ p_src_parameter) {
			if(!(p_parameter->p_s_name = p_s_Alloc(p_src_parameter->p_s_name)))
				return false;
			p_parameter->n_resource_index = p_src_parameter->n_resource_index;
			p_parameter->b_enviroment = p_src_parameter->b_enviroment;
		}
		// copy parameters

		const TProgram::TTexUnitInfo *p_src_tex_unit = p_low_src->p_parameter;
		for(TProgram::TTexUnitInfo *p_tex_unit = p_low->p_texture_unit,
		   *p_end = p_low->p_texture_unit + p_low->n_texture_unit_num; p_tex_unit < p_end;
		   ++ p_tex_unit, ++ p_src_tex_unit) {
			if(!(p_tex_unit->p_s_name = p_s_Alloc(p_src_tex_unit->p_s_name)))
				return false;
			p_tex_unit->n_resource_index = p_src_tex_unit->n_resource_index;
		}
		// copy texture units
	}
	// copy "low" section

	return true;
}

char *TShaderInfo::p_s_Alloc(const char *p_s_str)
{
	char *p_s_new;
	if(!(p_s_new = new char[strlen(p_s_str) + 1]))
		return 0;
	strcpy(p_s_new, p_s_str);
	return p_s_new;
}

/*
 *								=== ~TShaderInfo ===
 */

/*
 *								=== CGLShaderObject ===
 */

/*class CGLShaderObject { friend CShaderParameterRef;
protected:
	const TShaderInfo *m_p_vertex_shader;
	const TShaderInfo *m_p_fragment_shader;
	bool m_b_use_shader_target;

	CGLProgram *m_p_shader;

	std::vector<TParameterHandle> m_parameter_list;
	std::vector<CShaderParameterRef*> m_parameter_ref_list;

public:*/

/*
 *	CGLShaderObject::CGLShaderObject(const TShaderInfo *p_vertex_shader,
 *		const TShaderInfo *p_fragment_shader)
 *		- default constructor
 *		- m_p_vertex_shader and m_p_fragment_shader holds shader source code,
 *		  one of them can be 0
 *		- pointers are going to be referenced only, objects they point must remain
 *		  allocated till destruction of this object
 */
CGLShaderObject::CGLShaderObject(const TShaderInfo *p_vertex_shader,
	const TShaderInfo *p_fragment_shader)
	:m_p_vertex_shader(p_vertex_shader), m_p_fragment_shader(p_fragment_shader), m_p_shader(0)
{
	m_b_use_shader_target = b_ShaderAvailability(true) && b_ShaderSupport(true);
	// see it's possible to use high-level shading language
	// (in case it isn't, it still doesn't mean it's possible to use low-level shading language)
}

/*
 *	CGLShaderObject::~CGLShaderObject()
 *		- default destructor
 */
CGLShaderObject::~CGLShaderObject()
{
	CGLState cur_state(wglGetCurrentDC(), wglGetCurrentContext());
	cur_state.Update_States();
	Delete_Shaders(&cur_state);
}

/*
 *	const char *CGLShaderObject::p_s_VertexShader_Name()
 *		- returns name of vertex shader if present or 0 if not present
 */
const char *CGLShaderObject::p_s_VertexShader_Name()
{
	return (m_p_vertex_shader)? m_p_vertex_shader->p_s_name : 0;
}

/*
 *	const char *CGLShaderObject::p_s_FragmentShader_Name()
 *		- returns name of fragment shader if present or 0 if not present
 */
const char *CGLShaderObject::p_s_FragmentShader_Name()
{
	return (m_p_fragment_shader)? m_p_fragment_shader->p_s_name : 0;
}

/*
 *	bool CGLShaderObject::b_ShaderSupport(bool b_shader_target) const
 *		- returns true in case extensions necessary to compile supplied
 *		  shaders is available (in case vertex shader only was supplied,
 *		  vertex shader support is sufficient, the same applies for fragment shader)
 *		- b_shader_target specifies shader language (low-level program;
 *		  b_shader_target = false or high-level shader; b_shader_target = true)
 *		  note this function doesn't check for source code availability in
 *		  supplied shaders (i.e. it's possible low-level shading language is supported
 *		  but supplied shaders are available in high-level shading language only)
 *		- returns false in case (all) required extensions aren't supported
 */
bool CGLShaderObject::b_ShaderSupport(bool b_shader_target) const
{
	return (b_shader_target && (!m_p_vertex_shader || m_p_vertex_shader->p_high) &&
		(!m_p_fragment_shader || m_p_fragment_shader->p_high) &&
		(CGL_ARB_shader::b_Supported(m_p_vertex_shader != 0, m_p_fragment_shader != 0) ||
		CGL_Core_shader::b_Supported(m_p_vertex_shader != 0, m_p_fragment_shader != 0))) ||
		(!b_shader_target && (!m_p_vertex_shader || m_p_vertex_shader->p_low) &&
		(!m_p_fragment_shader || m_p_fragment_shader->p_low) &&
		CGL_ARB_program::b_Supported(m_p_vertex_shader != 0, m_p_fragment_shader != 0));
}

/*
 *	bool CGLShaderObject::b_ShaderAvailability(bool b_shader_target) const
 *		- returns true in case shaders in language, specified by b_shader_target are
 *		  available (returns false in case no shader was specified)
 *		- b_shader_target specifies shader language (low-level program;
 *		  b_shader_target = false or high-level shader; b_shader_target = true)
 *		  note this function doesn't check for OpenGL support
 *		- returns false in case (all specified) shaders are not available
 *		  in a given language
 */
bool CGLShaderObject::b_ShaderAvailability(bool b_shader_target) const
{
	return (m_p_vertex_shader || m_p_fragment_shader) &&
		((b_shader_target && (!m_p_vertex_shader || m_p_vertex_shader->p_high) &&
		(!m_p_fragment_shader || m_p_fragment_shader->p_high)) ||
		(!b_shader_target && (!m_p_vertex_shader || m_p_vertex_shader->p_low) &&
		(!m_p_fragment_shader || m_p_fragment_shader->p_low)));
}

/*
 *	bool CGLShaderObject::b_Status() const
 *		- returns true in case vertex and pixel shaders are compatible
 *		  so it's possible to compile them; otherwise returns false
 */
bool CGLShaderObject::b_Status() const
{
	return (b_ShaderAvailability(true) && b_ShaderSupport(true)) ||
		(b_ShaderAvailability(false) && b_ShaderSupport(false));
}

/*
 *	bool CGLShaderObject::b_ShaderTarget() const
 *		- returns true in case high level shading language is going to be used
 *		- returns false in case low level shading language is going to be used
 */
bool CGLShaderObject::b_ShaderTarget() const
{
	return m_b_use_shader_target;
}

/*
 *	void CGLShaderObject::Prefer_ShaderTarget(bool b_shader_target)
 *		- hint to use low (b_shader_target = false) or high (b_shader_target = true)
 *		  level shading language
 *		- in case shaders are present in desired shading  language, it is going
 *		  to be used. otherwise available shading language is going to be used
 */
void CGLShaderObject::Prefer_ShaderTarget(bool b_shader_target)
{
	if(b_ShaderAvailability(b_shader_target) && b_ShaderSupport(b_shader_target))
		m_b_use_shader_target = b_shader_target;
}

/*
 *	bool CGLShaderObject::Compile_Link(CGLState *p_state, char *&r_p_s_compile_log,
 *		char *&r_p_s_link_log)
 *		- compile and link shaders
 *		- p_state is active opengl state guard
 *		- r_p_s_compile_log is compilation log, (compiller error / warnings output)
 *		  r_p_s_link_log is linkage log, (linker error / warnings output)
 *		  in case compiller / linker generated error / warning (text) output,
 *		  string is allocated and pointer is put here. (it has to be freed) otherwise
 *		  the pointer is set to 0
 *		- returns true in case shaders were successfuly compiled and linked
 *		  (note despite that warnings could have been generated, see log string pointers)
 */
bool CGLShaderObject::Compile_Link(CGLState *p_state, char *&r_p_s_compile_log,
	char *&r_p_s_link_log)
{
	r_p_s_compile_log = 0;
	r_p_s_link_log = 0;

	if(!b_ShaderAvailability(m_b_use_shader_target) || !b_ShaderSupport(m_b_use_shader_target)) {
		m_b_use_shader_target = !m_b_use_shader_target;
		if(!b_ShaderAvailability(m_b_use_shader_target) || !b_ShaderSupport(m_b_use_shader_target))
			return false;
	}
	// see if all necessary shaders and extensions are available

	_ASSERTE(p_state);
	Delete_Shaders(p_state);

	if(m_b_use_shader_target) {
		if(CGL_Core_shader::b_Supported(m_p_vertex_shader != 0, m_p_fragment_shader != 0)) {
			if(!(m_p_shader = new CGL_Core_shader(
			   (m_p_vertex_shader)? m_p_vertex_shader->p_high->p_s_source_code : 0,
			   (m_p_fragment_shader)? m_p_fragment_shader->p_high->p_s_source_code : 0)))
				return false;
		} else /*if(CGL_ARB_shader::b_Supported(m_p_vertex_shader != 0,
		   m_p_fragment_shader != 0))*/ {
			_ASSERTE(CGL_ARB_shader::b_Supported(m_p_vertex_shader != 0, m_p_fragment_shader != 0));
			if(!(m_p_shader = new CGL_ARB_shader(
			   (m_p_vertex_shader)? m_p_vertex_shader->p_high->p_s_source_code : 0,
			   (m_p_fragment_shader)? m_p_fragment_shader->p_high->p_s_source_code : 0)))
				return false;
		}
	} else {
		_ASSERTE(CGL_ARB_program::b_Supported(m_p_vertex_shader != 0, m_p_fragment_shader != 0));
		if(!(m_p_shader = new CGL_ARB_program(
		   (m_p_vertex_shader)? m_p_vertex_shader->p_low->p_s_source_code : 0,
		   (m_p_fragment_shader)? m_p_fragment_shader->p_low->p_s_source_code : 0)))
			return false;
	}
	// create a new shader object

	if(!m_p_shader->Compile(p_state, r_p_s_compile_log) ||
	   (m_b_use_shader_target && !((CGLShader*)m_p_shader)->Link(r_p_s_link_log))) {
		delete m_p_shader;
		m_p_shader = 0;
		return false;
	}
	// compile and link shader

	if((m_b_use_shader_target && !Resolve_Uniforms()) ||
	   (!m_b_use_shader_target && !Fill_Parameters()))
		return false;
	// fill lists with shader parameter handles

	return true;
}

/*
 *	bool CGLShaderObject::Resolve_Uniforms()
 *		- fill lists with shader uniform handles
 *		- returns true on success, false on failure
 *		- assumes parameter handle lists are empty
 */
bool CGLShaderObject::Resolve_Uniforms()
{
	if(!m_b_use_shader_target || !m_p_shader)
		return false;

	int n_sampler_num = 0;
	int n_parameter_num = 0;
	int n_int_parameter_num = 0;
	int n_bool_parameter_num = 0;
	int n_matrix_parameter_num = 0;

	char **p_name_list;
	if(!(p_name_list = new char*[((CGLShader*)m_p_shader)->n_Uniform_Num()]))
		return false;
	int *p_content_type_list;
	if(!(p_content_type_list = new int[((CGLShader*)m_p_shader)->n_Uniform_Num()]))
		return false;

	for(int n_pass = 0; n_pass < 2; ++ n_pass) {
		for(int i = 0; i < ((CGLShader*)m_p_shader)->n_Uniform_Num(); ++ i) {
			if(!(p_name_list[i] = ((CGLShader*)m_p_shader)->p_s_Uniform_Name(i,
			   p_content_type_list[i]))) {
				for(int j = -- i; j >= 0; -- j)
					delete[] p_name_list[i];
				delete[] p_name_list;
				delete[] p_content_type_list;
				return false;
			}

			if(p_content_type_list[i] < shader_param_Float ||
			   p_content_type_list[i] > shader_sampler_2D_Rect_Shadow) {
				_ASSERTE(!n_pass);
				for(int j = i; j >= 0; -- j)
					delete[] p_name_list[i];
				delete[] p_name_list;
				delete[] p_content_type_list;
				return false;
				// error
			} else if(p_content_type_list[i] <= shader_param_Float4) {
				_ASSERTE(p_content_type_list[i] >= shader_param_Float &&
					p_content_type_list[i] <= shader_param_Float4);
				if(n_pass) {
					m_parameter_list.push_back(TVectorParamHandle(shader_param_Uniform,
						((CGLShader*)m_p_shader)->n_GetUniformLocation(p_name_list[i]),
						p_content_type_list[i], p_name_list[i]));
				} else
					n_parameter_num ++;
			} else if(p_content_type_list[i] <= shader_param_Int4) {
				_ASSERTE(p_content_type_list[i] >= shader_param_Int &&
					p_content_type_list[i] <= shader_param_Int4);
				if(n_pass) {
					m_int_parameter_list.push_back(TVectorParamHandle(shader_param_Uniform,
						((CGLShader*)m_p_shader)->n_GetUniformLocation(p_name_list[i]),
						p_content_type_list[i], p_name_list[i]));
				} else
					n_int_parameter_num ++;
			} else if(p_content_type_list[i] <= shader_param_Bool4) {
				_ASSERTE(p_content_type_list[i] >= shader_param_Bool &&
					p_content_type_list[i] <= shader_param_Bool4);
				if(n_pass) {
					m_bool_parameter_list.push_back(TVectorParamHandle(shader_param_Uniform,
						((CGLShader*)m_p_shader)->n_GetUniformLocation(p_name_list[i]),
						p_content_type_list[i], p_name_list[i]));
				} else
					n_bool_parameter_num ++;
			} else if(p_content_type_list[i] <= shader_param_Matrix4) {
				_ASSERTE(p_content_type_list[i] >= shader_param_Matrix2 &&
					p_content_type_list[i] <= shader_param_Matrix4);
				if(n_pass) {
					m_matrix_parameter_list.push_back(TMatrixParamHandle(shader_param_Uniform,
						((CGLShader*)m_p_shader)->n_GetUniformLocation(p_name_list[i]),
						p_content_type_list[i], p_name_list[i]));
				} else
					n_matrix_parameter_num ++;
			} else {
				_ASSERTE(p_content_type_list[i] >= shader_sampler_1D &&
					p_content_type_list[i] <= shader_sampler_2D_Rect_Shadow);
				if(n_pass) {
					m_sampler_list.push_back(TSamplerHandle(shader_param_Uniform,
						((CGLShader*)m_p_shader)->n_GetUniformLocation(p_name_list[i]),
						p_content_type_list[i], p_name_list[i]));
				} else
					n_sampler_num ++;
			}
		}

		if(n_pass) {
			delete[] p_name_list; // names are referenced by parameter handles
			delete[] p_content_type_list;
			// do not need this anymore
			break;
		}

		m_sampler_list.reserve(n_sampler_num);
		m_parameter_list.reserve(n_parameter_num);
		m_int_parameter_list.reserve(n_int_parameter_num);
		m_bool_parameter_list.reserve(n_bool_parameter_num);
		m_matrix_parameter_list.reserve(n_matrix_parameter_num);
		if(m_sampler_list.capacity() < n_sampler_num ||
		   m_parameter_list.capacity() < n_parameter_num ||
		   m_int_parameter_list.capacity() < n_int_parameter_num ||
		   m_bool_parameter_list.capacity() < n_bool_parameter_num ||
		   m_matrix_parameter_list.capacity() < n_matrix_parameter_num) {
			for(int i = 0; i < ((CGLShader*)m_p_shader)->n_Uniform_Num(); ++ i)
				delete[] p_name_list[i];
			delete[] p_name_list;
			delete[] p_content_type_list;
			return false;
		}
		// alloc space for all parameter handles
	}
	// get shader names and types (1st pass)
	// fill-in parameter handles (2nd pass)

	for(int i = 0; i < 2; ++ i) {
		const TShaderInfo *p_shader_info;
		if(!(p_shader_info = (i)? m_p_vertex_shader : m_p_fragment_shader))
			continue;
		TShaderInfo::TShader *p_shader = p_shader_info->p_high;
		for(TShaderInfo::TShader::TSamplerInfo *p_sampler = p_shader->p_sampler,
		   *p_end = p_shader->p_sampler + p_shader->n_sampler_num;
		   p_sampler < p_end; ++ p_sampler) {
			std::vector<TSamplerHandle>::iterator p_sam;
			if((p_sam = std::find_if(m_sampler_list.begin(), m_sampler_list.end(),
			   CParamHandleNameCompare(p_sampler->p_s_name))) == m_sampler_list.end())
				return false;
			(*p_sam).n_texture_unit = p_sampler->n_resource_index;
		}
	}
	// explicitly assign texture units

	for(std::vector<TSamplerHandle>::iterator p_sampler_handle =
	   std::find_if(m_sampler_list.begin(), m_sampler_list.end(),
	   std::mem_fun_ref(&TSamplerHandle::b_UnassignedTextureUnit));
	   p_sampler_handle != m_sampler_list.end(); p_sampler_handle =
	   std::find_if(p_sampler_handle + 1, m_sampler_list.end(),
	   std::mem_fun_ref(&TSamplerHandle::b_UnassignedTextureUnit))) {
		if(p_sampler_handle == m_sampler_list.begin())
			(*p_sampler_handle).n_texture_unit = 0;
		else
			(*p_sampler_handle).n_texture_unit = (*(p_sampler_handle - 1)).n_texture_unit + 1;
	}
	// assign unassigned texture units

	return true;
}

/*
 *	bool CGLShaderObject::Resolve_Uniforms()
 *		- fill lists with program parameter handles
 *		- returns true on success, false on failure
 *		- assumes parameter handle lists are empty
 */
bool CGLShaderObject::Fill_Parameters()
{
	if(m_b_use_shader_target || !m_p_shader)
		return false;

	int n_parameter_num = ((m_p_vertex_shader)? m_p_vertex_shader->p_low->n_parameter_num : 0) +
		((m_p_fragment_shader)? m_p_fragment_shader->p_low->n_parameter_num : 0);
	if(m_parameter_list.capacity() < n_parameter_num)
		return false;
	// alloc space for all parameter handles

	for(int i = 0; i < 2; ++ i) {
		const TShaderInfo *p_shader_info;
		if(!(p_shader_info = (i)? m_p_vertex_shader : m_p_fragment_shader))
			continue;
		TShaderInfo::TProgram *p_program = p_shader_info->p_low;
		for(TShaderInfo::TProgram::TParamInfo *p_parameter = p_program->p_parameter,
		   *p_end = p_program->p_parameter + p_program->n_parameter_num;
		   p_parameter < p_end; ++ p_parameter) {
			char *p_s_name;
			if(!(p_s_name = new char[strlen(p_parameter->p_s_name) + 1]))
				return false;
			strcpy(p_s_name, p_parameter->p_s_name);
			// copy name
			
			int n_target = (p_parameter->b_enviroment)? shader_param_Parameter_Env :
				((p_shader_info == m_p_vertex_shader)? shader_param_Parameter_Local_VP :
				shader_param_Parameter_Local_FP);
			// find out what seems to be parameter target

			m_bool_parameter_list.push_back(TVectorParamHandle(n_target,
				p_parameter->n_resource_index, shader_param_Float4, p_s_name));
		}
	}
	// add all parameter handles

	return true;
}

/*
 *	CShaderVectorParameterRef<float> *CGLShaderObject::p_Find_Parameter(
 *		const char *p_s_parameter_name)
 *		- find float shader parameter by name p_s_parameter_name
 *		- returns pointer to shader parameter object on success, 0 on failure
 *		  (in case shaders weren't compiled or parameter with p_s_parameter_name doesn't exist)
 */
CShaderVectorParameterRef<float> *CGLShaderObject::p_Find_Parameter(const char *p_s_parameter_name)
{
	if(!m_p_shader)
		return 0;

	std::vector<TVectorParamHandle>::iterator p_param_handle;
	if((p_param_handle = std::find_if(m_parameter_list.begin(),
	   m_parameter_list.end(), CParamHandleNameCompare(p_s_parameter_name))) ==
	   m_parameter_list.end())
		return 0;
	// find uniform handle

	_ASSERTE((*p_param_handle).n_type >= shader_param_Float &&
		(*p_param_handle).n_type <= shader_param_Float4);

	return new CShaderVectorParameterRef<float>((*p_param_handle).t_value.p_float_value,
		(*p_param_handle).n_type - (int)shader_param_Float + 1, this);
}

/*
 *	CShaderVectorParameterRef<int> *CGLShaderObject::p_Find_IntParameter(
 *		const char *p_s_parameter_name)
 *		- find integer shader parameter by name p_s_parameter_name
 *		- returns pointer to shader parameter object on success, 0 on failure
 *		  (in case shaders weren't compiled or parameter with p_s_parameter_name doesn't exist)
 *		- note integer shader parameters are available in high-level shading language only
 */
CShaderVectorParameterRef<int> *CGLShaderObject::p_Find_IntParameter(const char *p_s_parameter_name)
{
	if(!m_p_shader || !m_b_use_shader_target)
		return 0;

	std::vector<TVectorParamHandle>::iterator p_param_handle;
	if((p_param_handle = std::find_if(m_int_parameter_list.begin(),
	   m_int_parameter_list.end(), CParamHandleNameCompare(p_s_parameter_name))) ==
	   m_int_parameter_list.end())
		return 0;
	// find uniform handle

	_ASSERTE((*p_param_handle).n_type >= shader_param_Int &&
		(*p_param_handle).n_type <= shader_param_Int4);

	return new CShaderVectorParameterRef<int>((*p_param_handle).t_value.p_int_value,
		(*p_param_handle).n_type - (int)shader_param_Int + 1, this);
}

/*
 *	CShaderVectorParameterRef<int> *CGLShaderObject::p_Find_BoolParameter(
 *		const char *p_s_parameter_name)
 *		- find boolean shader parameter by name p_s_parameter_name
 *		- returns pointer to shader parameter object on success, 0 on failure
 *		  (in case shaders weren't compiled or parameter with p_s_parameter_name doesn't exist)
 *		- note boolean shader parameters are available in high-level shading language only
 */
CShaderVectorParameterRef<int> *CGLShaderObject::p_Find_BoolParameter(
	const char *p_s_parameter_name)
{
	if(!m_p_shader || !m_b_use_shader_target)
		return 0;

	std::vector<TVectorParamHandle>::iterator p_param_handle;
	if((p_param_handle = std::find_if(m_bool_parameter_list.begin(),
	   m_bool_parameter_list.end(), CParamHandleNameCompare(p_s_parameter_name))) ==
	   m_bool_parameter_list.end())
		return 0;
	// find uniform handle

	_ASSERTE((*p_param_handle).n_type >= shader_param_Bool &&
		(*p_param_handle).n_type <= shader_param_Bool4);

	return new CShaderVectorParameterRef<int>((*p_param_handle).t_value.p_int_value,
		(*p_param_handle).n_type - (int)shader_param_Int + 1, this);
}

/*
 *	CShaderMatrixParameterRef *CGLShaderObject::p_Find_MatrixParameter(
 *		const char *p_s_parameter_name)
 *		- find matrix shader parameter by name p_s_parameter_name
 *		- returns pointer to shader parameter object on success, 0 on failure
 *		  (in case shaders weren't compiled or parameter with p_s_parameter_name doesn't exist)
 *		- note matrix shader parameters are available in high-level shading language only
 */
CShaderMatrixParameterRef *CGLShaderObject::p_Find_MatrixParameter(const char *p_s_parameter_name)
{
	if(!m_p_shader || !m_b_use_shader_target)
		return 0;

	std::vector<TMatrixParamHandle>::iterator p_param_handle;
	if((p_param_handle = std::find_if(m_matrix_parameter_list.begin(),
	   m_matrix_parameter_list.end(), CParamHandleNameCompare(p_s_parameter_name))) ==
	   m_matrix_parameter_list.end())
		return 0;
	// find uniform handle

	_ASSERTE((*p_param_handle).n_type >= shader_param_Matrix2 &&
		(*p_param_handle).n_type <= shader_param_Matrix4);

	return new CShaderMatrixParameterRef((*p_param_handle).p_value,
		(*p_param_handle).n_type - (int)shader_param_Matrix2 + 2, this);
}

/*
 *	int CGLShaderObject::n_TextureType(const char *p_s_sampler_name) const
 *		- find texture type by name p_s_sampler_name
 *		- returns sampler type (one of shader_sampler_1D, shader_sampler_2D, shader_sampler_3D,
 *		  shader_sampler_Cube, shader_sampler_1D_Shadow, shader_sampler_2D_Shadow,
 *		  shader_sampler_2D_Rect, shader_sampler_2D_Rect_Shadow) or 0 on failure
 *		- note samplers are available in high-level shading language only
 */
int CGLShaderObject::n_TextureType(const char *p_s_sampler_name) const
{
	if(!m_p_shader || !m_b_use_shader_target)
		return 0;

	std::vector<TSamplerHandle>::iterator p_sam;
	if((p_sam = std::find_if((std::vector<TSamplerHandle>::iterator)m_sampler_list.begin(),
	   (std::vector<TSamplerHandle>::iterator)m_sampler_list.end(),
	   CParamHandleNameCompare(p_s_sampler_name))) == m_sampler_list.end())
		return 0;

	return (*p_sam).n_type;
}

/*
 *	int CGLShaderObject::n_TextureUnit(const char *p_s_sampler_name) const
 *		- find texture unit by name p_s_sampler_name
 *		- returns texture unit index (0 to GL_MAX_TEXTURE_UNITS - 1) or -1 on failure
 */
int CGLShaderObject::n_TextureUnit(const char *p_s_sampler_name) const
{
	if(!m_p_shader)
		return -1;

	if(m_b_use_shader_target) {
		std::vector<TSamplerHandle>::iterator p_sam;
		if((p_sam = std::find_if((std::vector<TSamplerHandle>::iterator)m_sampler_list.begin(),
		   (std::vector<TSamplerHandle>::iterator)m_sampler_list.end(),
		   CParamHandleNameCompare(p_s_sampler_name))) == m_sampler_list.end())
			return -1;
		return (*p_sam).n_texture_unit;
	} else {
		for(int i = 0; i < 2; ++ i) {
			const TShaderInfo *p_shader_info;
			if(!(p_shader_info = (i)? m_p_vertex_shader : m_p_fragment_shader))
				continue;
			TShaderInfo::TProgram *p_program = p_shader_info->p_low;
			for(TShaderInfo::TProgram::TTexUnitInfo *p_tex_unit = p_program->p_texture_unit,
			   *p_end = p_program->p_texture_unit + p_program->n_texture_unit_num;
			   p_tex_unit < p_end; ++ p_tex_unit) {
				if(!strcmp(p_tex_unit->p_s_name, p_s_sampler_name))
					return p_tex_unit->n_resource_index;
			}
		}
		// go trough all texture unit names
		return -1;
	}
}

/*
 *	bool CGLShaderObject::Bind(CGLState *p_state) const
 *		- bind shader, set supplied parameters
 */
bool CGLShaderObject::Bind(CGLState *p_state) const
{
	if(!m_p_shader)
		return false;
	if(m_p_shader->n_Api() == shader_api_Core_shader) {
		CGL_Core_shader *p_shader = (CGL_Core_shader*)m_p_shader;

		p_shader->_Bind(p_state);
		// bind shader

		if(!std::for_each(m_sampler_list.begin(), m_sampler_list.end(),
		   CSamplerBinder<CGL_Core_shader>(p_state, p_shader)))
			return false;
		// set all sampler locations

		if(!std::for_each(m_parameter_list.begin(), m_parameter_list.end(),
		   CVectorParamBinder_Float<CGL_Core_shader>(p_state, p_shader)))
			return false;
		// set all float-vector parameter locations

		if(!std::for_each(m_int_parameter_list.begin(), m_int_parameter_list.end(),
		   CVectorParamBinder_Int<CGL_Core_shader>(p_state, p_shader)))
			return false;
		// set all int-vector parameter locations

		if(!std::for_each(m_bool_parameter_list.begin(), m_bool_parameter_list.end(),
		   CVectorParamBinder_Bool<CGL_Core_shader>(p_state, p_shader)))
			return false;
		// set all bool-vector parameter locations

		if(!std::for_each(m_matrix_parameter_list.begin(), m_matrix_parameter_list.end(),
		   CMatrixParamBinder<CGL_Core_shader>(p_state, p_shader)))
			return false;
		// set all float-matrix parameter locations
	} else if(m_p_shader->n_Api() == shader_api_ARB_shader) {
		CGL_ARB_shader *p_shader = (CGL_ARB_shader*)m_p_shader;

		p_shader->_Bind(p_state);
		// bind shader

		if(!std::for_each(m_sampler_list.begin(), m_sampler_list.end(),
		   CSamplerBinder<CGL_ARB_shader>(p_state, p_shader)))
			return false;
		// set all sampler locations

		if(!std::for_each(m_parameter_list.begin(), m_parameter_list.end(),
		   CVectorParamBinder_Float<CGL_ARB_shader>(p_state, p_shader)))
			return false;
		// set all float-vector parameter locations

		if(!std::for_each(m_int_parameter_list.begin(), m_int_parameter_list.end(),
		   CVectorParamBinder_Int<CGL_ARB_shader>(p_state, p_shader)))
			return false;
		// set all int-vector parameter locations

		if(!std::for_each(m_bool_parameter_list.begin(), m_bool_parameter_list.end(),
		   CVectorParamBinder_Bool<CGL_ARB_shader>(p_state, p_shader)))
			return false;
		// set all bool-vector parameter locations

		if(!std::for_each(m_matrix_parameter_list.begin(), m_matrix_parameter_list.end(),
		   CMatrixParamBinder<CGL_ARB_shader>(p_state, p_shader)))
			return false;
		// set all float-matrix parameter locations
	} else if(m_p_shader->n_Api() == shader_api_ARB_program) {
		CGL_ARB_program *p_program = (CGL_ARB_program*)m_p_shader;

		p_program->_Bind(p_state);
		// bind shader

		if(!std::for_each(m_parameter_list.begin(), m_parameter_list.end(),
		   CVectorParamBinder_Float4<CGL_ARB_program>(p_state, p_program)))
			return false;
		// set all float-vector parameter locations
	} else
		return false;

	return true;
}

/*
 *	void CGLShaderObject::Release(CGLState *p_state) const
 *		- release shader
 */
void CGLShaderObject::Release(CGLState *p_state) const
{
	if(m_p_shader)
		m_p_shader->Release(p_state);
}

/*
 *	int CGLShaderObject::n_Api() const
 *		- return id of api being used (shader_api_ARB_program,
 *		  shader_api_ARB_shader or shader_api_Core_shader)
 */
int CGLShaderObject::n_Api() const
{
	if(!m_p_shader)
		return shader_api_None;
	return m_p_shader->n_Api();
}

/*
 *	void CGLShaderObject::Delete_Shaders(CGLState *p_state)
 *		- delete opengl shader objects (in case they were created)
 */
void CGLShaderObject::Delete_Shaders(CGLState *p_state)
{
	if(m_p_shader) {
		m_p_shader->GL_Free(p_state);
		delete m_p_shader;
		m_p_shader = 0;
	}
	// delete shader

	m_parameter_list.clear();
	m_sampler_list.clear();
	m_int_parameter_list.clear();
	m_bool_parameter_list.clear();
	m_matrix_parameter_list.clear();
	// remove parameter handles

	std::for_each(m_parameter_ref_list.begin(), m_parameter_ref_list.end(),
		std::mem_fun(&CShaderParameterRef<float>::Invalidate));
	m_parameter_ref_list.clear();
	std::for_each(m_int_bool_parameter_ref_list.begin(), m_int_bool_parameter_ref_list.end(),
		std::mem_fun(&CShaderParameterRef<int>::Invalidate));
	m_int_bool_parameter_ref_list.clear();
	// remove and invalidate parameter references
}

/*
 *	void CGLShaderObject::Remove_Parameter(CShaderParameterRef<float> *p_parameter,
 *		bool b_invalidate = true)
 *		- remove parameter p_parameter from the list
 *		- in case b_invalidate is true, p_parameter->Invalidate() is going to be called
 *		  (to avoid infinite recurse and "zombie" parameters)
 */
void CGLShaderObject::Remove_Parameter(CShaderParameterRef<float> *p_parameter, bool b_invalidate)
{
	std::vector<CShaderParameterRef<float>*>::iterator p_iter =
		std::find(m_parameter_ref_list.begin(), m_parameter_ref_list.end(), p_parameter);
	if(p_iter != m_parameter_ref_list.end())
		m_parameter_ref_list.erase(p_iter);
	if(b_invalidate)
		p_parameter->Invalidate();
}

/*
 *	void CGLShaderObject::Remove_Parameter(CShaderParameterRef<int> *p_parameter,
 *		bool b_invalidate = true)
 *		- remove parameter p_parameter from the list
 *		- in case b_invalidate is true, p_parameter->Invalidate() is going to be called
 *		  (to avoid infinite recurse and "zombie" parameters)
 */
void CGLShaderObject::Remove_Parameter(CShaderParameterRef<int> *p_parameter, bool b_invalidate)
{
	std::vector<CShaderParameterRef<int>*>::iterator p_iter =
		std::find(m_int_bool_parameter_ref_list.begin(),
		m_int_bool_parameter_ref_list.end(), p_parameter);
	if(p_iter != m_int_bool_parameter_ref_list.end())
		m_int_bool_parameter_ref_list.erase(p_iter);
	if(b_invalidate)
		p_parameter->Invalidate();
}

/*
 *								=== ~CGLShaderObject ===
 */
